import sys, re, os, copy
import numpy as np
import numpy.linalg as lin
import io as strio
import itertools as itr
from numpy import sin, cos

prec = "14.8f"
real_fmt = "{{:{}}}".format(prec)
xyz_fmt = "".join(["{{r[{}]:{}}}".format(i, prec) for i in range(3)])
fmt_library = {"aims": "{name:<15s}" + xyz_fmt,
               "aims_elem": "{name:<15s}" + xyz_fmt + 4*" " + "{elem:<2s}",
               "xyz": "{elem:<2s}" + xyz_fmt,
               "vasp": "  " + xyz_fmt}

def split(l):
    return re.split(r"\s+", l.strip())

def mass(elem):
    data = [1.0079, 4.0026, 6.941, 9.0122, 10.811, 12.0107, 14.0067, 15.9994,
            18.9984, 20.1797, 22.9897, 24.305, 26.9815, 28.0855, 30.9738, 32.065,
            35.453, 39.948, 39.0983, 40.078, 44.9559, 47.867, 50.9415, 51.9961,
            54.938, 55.845, 58.9332, 58.6934, 63.546, 65.39, 69.723, 72.64,
            74.9216, 78.96, 79.904, 83.8, 85.4678, 87.62, 88.9059, 91.224,
            92.9064, 95.94, 98, 101.07, 102.9055, 106.42, 107.8682, 112.411,
            114.818, 118.71, 121.76, 127.6, 126.9045, 131.293, 132.9055, 137.327,
            138.9055, 140.116, 140.9077, 144.24, 145, 150.36, 151.964, 157.25,
            158.9253, 162.5, 164.9303, 167.259, 168.9342, 173.04, 174.967, 178.49,
            180.9479, 183.84, 186.207, 190.23, 192.217, 195.078, 196.9665, 200.59,
            204.3833, 207.2, 208.9804, 209, 210, 222, 223, 226,
            227, 232.0381, 231.0359, 238.0289]
    try:
        return data[elem-1]
    except TypeError:
        return data[element(elem)-1]

def covalent_radius(elem):
    data = [0.38, 0.32, 1.34, 0.9, 0.82, 0.77, 0.75, 0.73, 0.71, 0.69, 1.54, 1.3, 1.18, 
            1.11, 1.06, 1.02, 0.99, 0.97, 1.96, 1.74, 1.44, 1.36, 1.25, 1.27, 1.39, 1.25, 
            1.26, 1.21, 1.38, 1.31, 1.26, 1.22, 1.19, 1.16, 1.14, 1.1, 2.11, 1.92, 1.62, 1.48, 
            1.37, 1.45, 1.56, 1.26, 1.35, 1.31, 1.53, 1.48, 1.44, 1.41, 1.38, 1.35, 1.33, 1.3, 
            2.25, 1.98, 1.69, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.6, 1.5, 1.38, 1.46, 
            1.59, 1.28, 1.37, 1.28, 1.44, 1.49, 1.48, 1.47, 1.46, 0, 0, 1.45, 0, 0, 0, 0, 0, 0]
    try:
        r = data[elem-1]
    except TypeError:
        r = data[element(elem)-1]
    assert r > 0
    return r
    
def element(elem):
    data = ["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", 
            "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", 
            "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", 
            "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y", "Zr", 
            "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", 
            "Sb", "Te", "I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", 
            "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", 
            "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", 
            "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", 
            "Pa", "U"]
    try: 
        return data[elem-1] 
    except TypeError: 
        return data.index(elem.capitalize())+1

def connectedness(C):
    n = C.shape[0]
    found = -np.ones(n, int)
    queue = np.zeros(n, int)
    ifrag = -1
    for i in range(n):
        if found[i] >= 0:
            continue
        ifrag += 1
        queue[0] = i
        a = 0
        b = 1
        while b-a > 0:
            node = queue[a]
            a += 1
            found[node] = ifrag
            succs = np.flatnonzero(C[node, :])
            for succ in succs:
                if found[succ] >= 0 or (queue[a:b] == succ).any():
                    continue
                b += 1
                queue[b-1] = succ
    fragments = []
    for i in range(ifrag+1):
        fragments.append(np.flatnonzero(found == i).tolist())
    return fragments

class Atom:
    def __init__(self, Z, R):
        Z = element(Z) if isinstance(Z, str) else Z
        self.Z = Z
        self.R = np.array(R, float)
    def __str__(self):
        return self.to_str(fmt_library["xyz"])
    def copy(self):
        return copy.deepcopy(self)
    def to_str(self, fmt, data=None, elem_symb=True):
        data = data if data else {}
        data["elem"] = (element if elem_symb else str)(self.Z)
        data["r"] = self.R
        return fmt.format(**data)
    def dist(self, atom):
        return np.sqrt(sum((self.R-atom.R)**2))

class Molecule:
    def __init__(self, atoms=None, desc=None):
        self.atoms = atoms if atoms else []
        self.desc = desc if desc else ""
    def __str__(self):
        return self.to_str("xyz")
    @classmethod
    def from_str(cls, s, fmt="xyz"):
        ls = s.strip().split("\n")
        if fmt == "xyz":
            n = int(ls[0])
            desc = ls[1]
            atoms = []
            for l in ls[2:2+n]:
                t = split(l)
                atoms.append(Atom(t[0], list(map(float, t[1:4]))))
        elif fmt == "aims":
            atoms = []
            for l in ls:
                t = split(l)
                atoms.append(Atom(t[4], list(map(float, t[1:4]))))
        return cls(atoms, desc)
    def copy(self):
        return copy.deepcopy(self)
    def to_str(self, fmt):
        s = strio.StringIO()
        if fmt == "xyz":
            print(self.n(), file=s)
            print(self.desc, file=s)
            for a in self.atoms:
                print(a.to_str(fmt_library["xyz"]), file=s)
        elif fmt == "aims":
            if self.desc:
                print("# {}".format(self.dec), file=s)
            for a in self.atoms:
                print(a.to_str(fmt_library["aims_elem"], data={"name": "atom"}), file=s)
        return s.getvalue().rstrip()
    def n(self):
        return len(self.atoms)
    def mass(self):
        return sum([mass(a.Z) for a in self.atoms])
    def center_of_mass(self):
        return sum([mass(a.Z)*a.R for a in self.atoms])/self.mass()
    def shift(self, where=None, by=None):
        m = self.copy()
        if by:
            shift = by
        elif where:
            shift = np.array(where)-self.center_of_mass()
        assert not shift is None
        for a in m.atoms:
            a.R += shift
        return m
    def rotate(self, axis, phi, center=None):
        m = self.copy()
        phi = phi*np.pi/180
        rotmat = np.array([1,        0,         0,
                           0, cos(phi), -sin(phi),
                           0, sin(phi),  cos(phi)]).reshape(3, 3)
        shift = {"x": 0, "y": 1, "z": 2}[axis]
        for i_dim in [0, 1]:
            rotmat = np.roll(rotmat, shift, i_dim)
        center = np.array(center) if center else self.center_of_mass()
        for a in m.atoms:
            a.R = center+np.dot(rotmat, a.R-center)
        return m
    def find_fragments(self, scale=1.3):
        bond = np.zeros((self.n(), self.n()), bool)
        for i, a_i in enumerate(self.atoms):
            for j, a_j in enumerate(self.atoms):
                if i == j:
                    continue
                R_cov = covalent_radius(a_i.Z)+covalent_radius(a_j.Z)
                bond[i, j] = a_i.dist(a_j) < scale*R_cov
        frag_defs = connectedness(bond)
        fragments = []
        for frag in frag_defs:
            fragments.append(Molecule([self.atoms[i].copy() for i in frag]))
        return fragments
    def join(self, mol):
        return Molecule(self.copy().atoms + mol.copy().atoms)

class Crystal:
    def __init__(self, atoms=None, lattice_vectors=None, ortho=None, 
                 hexa=None, desc=None, coord_type="c"):
        self.desc = desc if desc else ""
        self.atoms = atoms if atoms else []
        if not ortho is None:
            self.lattice_vectors = np.diag(ortho)
        elif not hexa is None:
            self.lattice_vectors = np.array([(hexa[0], 0, 0),
                                             (hexa[1]/2, hexa[1]*np.sqrt(3)/2, 0),
                                             (0, 0, hexa[2])])
        else:
            self.lattice_vectors = lattice_vectors
        assert not self.lattice_vectors is None
        self.coord_type = coord_type
    def __str__(self):
        return self.to_str("aims")
    @classmethod
    def from_molecule(cls, mol, ortho=None, hexa=None):
        return cls(mol.copy().atoms, ortho=ortho, hexa=hexa, desc=mol.desc)
    @classmethod
    def from_str(cls, s, fmt):
        ls = s.strip().split("\n")
        if fmt == "vasp":
            desc = ls[0]
            scale = float(ls[1])
            lattice_vectors = scale*np.array([list(map(float, split(l))) for l in ls[2:5]])
            Zs = list(map(element, split(ls[5])))
            n_elems = list(map(int, split(ls[6])))
            n_atoms = sum(n_elems)
            Zs = list(itr.chain(*[itr.repeat(e, n) for e, n in zip(Zs, n_elems)]))
            coord_type = ls[7].strip().lower()
            if scale != 1:
                assert coord_type == "d"
            coords = np.array([list(map(float, split(l))) for l in ls[8:8+n_atoms]])
            atoms = [Atom(Z, R) for Z, R in zip(Zs, coords)]
        elif fmt == "aims":
            desc = ""
            lattice_vectors = []
            atoms = []
            coord_type = None
            for l in ls:
                t = split(l)
                if t[0] == "lattice_vector":
                    lattice_vectors.append(list(map(float, t[1:4])))
                elif t[0] in ["atom", "atom_frac"]:
                    coord_type_current = {"atom": "c", "atom_frac": "d"}[t[0]]
                    if coord_type:
                        assert coord_type == coord_type_current
                    else:
                        coord_type = coord_type_current
                    atoms.append(Atom(t[4], list(map(float, t[1:4]))))
            lattice_vectors = np.array(lattice_vectors)
        return cls(atoms, lattice_vectors, desc=desc, coord_type=coord_type)
    def set_coord_type(self, coord_type):
        if coord_type == self.coord_type:
            return
        elif coord_type == "c":
            self.coord_type = "c"
            for a in self.atoms:
                a.R = a.R.dot(self.lattice_vectors)
        elif coord_type == "d":
            self.coord_type = "d"
            for a in self.atoms:
                a.R = lin.solve(self.lattice_vectors.T, a.R)
    def copy(self):
        return copy.deepcopy(self)
    def to_str(self, fmt):
        s = strio.StringIO()
        if fmt == "vasp":
            print(self.desc, file=s)
            print(real_fmt.format(1), file=s)
            for vector in self.lattice_vectors:
                print(fmt_library["vasp"].format(r=vector), file=s)
            Zs = list(set([a.Z for a in self.atoms]))
            print(" ".join(map(element, Zs)), file=s)
            atoms = [[] for _ in Zs]
            for a in self.atoms:
                atoms[Zs.index(a.Z)].append(a)
            n_elems = list(map(len, atoms))
            atoms = list(itr.chain(*atoms))
            print(" ".join(map(str, n_elems)), file=s)
            print({"d": "direct", "c": "cartesian"}[self.coord_type], file=s)
            for a in atoms:
                print(a.to_str(fmt_library["vasp"]), file=s)
        elif fmt == "aims":
            if self.desc:
                print("# {}".format(self.desc), file=s)
            for vector in self.lattice_vectors:
                print(fmt_library["aims"].format(name="lattice_vector", r=vector), file=s)
            for a in self.atoms:
                name = {"c": "atom", "d": "atom_frac"}[self.coord_type]
                print(a.to_str(fmt_library["aims_elem"], data={"name": name}), file=s)
        return s.getvalue().rstrip()
    def to_molecule(self):
        return Molecule(self.copy().atoms)
    def n(self):
        return len(self.atoms)
    def copy_cell(self, dim):
        dim = np.array(dim)
        coord_type = self.coord_type
        self.set_coord_type("c")
        atoms = []
        for i in range(dim[0]):
            for j in range(dim[1]):
                for k in range(dim[2]):
                    shift = np.array((i, j, k))
                    for a in self.atoms:
                        a = a.copy()
                        a.R += shift.dot(self.lattice_vectors)
                        atoms.append(a)
        lattice_vectors = np.diag(dim).dot(self.lattice_vectors)
        self.set_coord_type(coord_type)
        c = Crystal(atoms, lattice_vectors, coord_type="c")
        c.set_coord_type(coord_type)
        self.set_coord_type(coord_type)
        return c
