#!/usr/bin/python
import os, sys, re, subprocess, shutil, glob, csv, io
import collections as col
import subprocess as subp
import numpy as np

class AIMS_output:
    def __init__(self, f):
        assert isinstance(f, io.TextIOBase)
        self.f = f
        self.results = {"MBD": {}, "Energies": {}}
        self.converged = False
        self.to_get = {"Number of atoms": self.get_atoms,
                       "Self-consistency cycle converged": self.set_converged,
                       "Performing Hirshfeld analysis": self.get_hirsh,
                       "Many-Body Dispersion": self.get_mbd,
                       "Total energy components": self.get_energy,
                       "decomposition of the XC Energy": self.get_xc}
        while self.readline():
            for key in self.to_get:
                if key in self.line:
                     self.to_get[key](key)
                     break
    def readline(self):
        self.line = self.f.readline()
        return self.line
    def get_atoms(self, key_to_get):
        n, = re.search(r"Number of atoms\s+\:\s+(\d+)", self.line).groups()
        n = int(n)
        self.n_atoms = n
        self.results["n_atoms"] = n
        del self.to_get[key_to_get]
    def set_converged(self, key_to_get):
        self.converged = True
        del self.to_get[key_to_get]
    def get_hirsh(self, key_to_get):
        atoms = []
        for i_atom in range(self.n_atoms):
            atom = {}
            self.readline()
            atom["Element"], = re.search(r"Atom\s+\d+\:\s+(\w+)", self.readline()).groups()
            for i_line in range(5):
                key, val = re.search(r"\|\s+(.*?)\s*\:(.*)", self.readline()).groups()
                key = key.strip()
                val = list(map(float, re.split(r"\s+", val.strip())))
                if len(val) == 1:
                    val = val[0]
                atom[key] = val
            for i_line in range(3):
                self.readline()
            atoms.append(atom)
        self.results["Hirshfeld"] = atoms
	#print (atoms)
        del self.to_get[key_to_get]
    def get_mbd(self, key_to_get):
        results = {}
        name, = re.search(r"Many-Body Dispersion \((.*)\) energy", self.line).groups()
        self.readline()
        if re.search(r"Dynamic.*polarizability", self.line):
            self.readline()
            labels, = re.search(r"\|\s+(\S.*)", self.readline()).groups()
            labels = re.split(r"\s+", labels.strip())
            val = [[] for label in labels]
            while not "---" in self.readline():
                r, = re.search(r"\|\s+(\S.*)", self.line).groups()
                r = list(map(float, re.split(r"\s+", r.strip())))
                for x, arr in zip(r, val):
                    arr.append(x)
            omega = dict((k, np.array(v)) for k, v in zip(labels, val))
            omega["integration_weight"] = np.array([0.0000000,
                        0.0786611, 0.0796400, 0.0816475, 0.0847872, 0.0892294,
                        0.0952317, 0.1031720, 0.1136050, 0.1273500, 0.1456520,
                        0.1704530, 0.2049170, 0.2544560, 0.3289620, 0.4480920,
                        0.6556060, 1.0659600, 2.0635700, 5.6851000, 50.955800])
            results["Dynamic molecular polarizability"] = omega
	    #print (omega)
            #for i_line in range(2):
            #    self.readline()
            #    print (self.readline())
            #atoms = []
            #for i_atom in range(self.n_atoms):
            #    elem, c6, alpha = re.search(r"\|\s+ATOM\s+\d+\s+(\w+)\s+(\S+)\s+(\S+)", self.readline()).groups()
            #    c6, alpha = float(c6), float(alpha)
            #    atoms.append((elem, c6, alpha))
            #results["Partitioned C6"] = atoms
            self.readline()
        if name != "MBD@rsSCS":
            while not "---" in self.readline():
                try:
                    key, val = re.search(r"\|\s+(.*?)\s*\:\s+([-\.\d]+)\s+Ha", self.line).groups()
                except AttributeError:
                    continue
                val = float(val)
                self.results["Energies"][key] = val
        self.results["MBD"][name] = results
    def get_energy(self, key_to_get):
        if not self.converged:
            return
        enes = self.results["Energies"]
        while "|" in self.readline():
            try:
                key, val = re.search(r"\|\s+(.*?)\s*\:\s+([-\.\d]+)\s+Ha", self.line).groups()
            except AttributeError:
                continue
            if key in ["Total energy", "MBD@rsSCS energy", "van der Waals energy corr."]:
                val = float(val)
                if key == "van der Waals energy corr.":
                    key = "TS energy"
                enes[key] = val
        if "MBD@rsSCS energy" in enes:
            vdw = "MBD@rsSCS energy"
        elif "TS energy" in enes:
            vdw = "TS energy"
        else:
            vdw = None
        if vdw:
            enes["DFT energy"] = enes["Total energy"]-enes[vdw]
            enes["Total energy + " + vdw] = enes["Total energy"]
            del enes["Total energy"]
        del self.to_get[key_to_get]
    def get_xc(self, key_to_get):
        while not "End decomposition" in self.readline():
            try:
                key, val = re.search(r"\s*(.*?)\s*\:\s+([-\.\d]+)\s+Ha", self.line).groups()
            except AttributeError:
                continue
            val = float(val)
            if key == "X Energy":
                self.results["Energies"]["Exchange energy"] = val
            if key == "C Energy GGA":
                self.results["Energies"]["Correlation energy"] = val


if __name__ == "__main__":
    import pprint
    pp = pprint.PrettyPrinter()
    with open(sys.argv[1]) as f:
        pp.pprint(AIMS_output(f).results)
